"""
Tests brewblox_emitter.relay
"""

import asyncio
import json

import pytest
from aiohttp.client_exceptions import ClientPayloadError
from brewblox_service import scheduler

from brewblox_emitter import relay

TESTED = relay.__name__


@pytest.fixture
def listener_mock(mocker):
    call_mock = mocker.patch(TESTED + '.events.get_listener')
    return call_mock.return_value


@pytest.fixture
async def app(app, mocker, listener_mock):
    mocker.patch(TESTED + '.CLEANUP_INTERVAL_S', 0.0001)

    scheduler.setup(app)
    relay.setup(app)
    return app


async def test_add_queue(app, client):
    rl = relay.get_relay(app)

    await rl._on_event_message(None, 'test', {
        'key': 'testkey',
        'duration': '10s',
        'data': {'stuff': True}
    })
    q1 = asyncio.Queue()
    q2 = asyncio.Queue()

    await rl.add_queue(q1)
    await rl.add_queue(q2)
    expected = {'testkey': {'stuff': True}}
    assert q1.get_nowait() == expected
    assert q2.get_nowait() == expected


async def test_expire(app, client):
    rl = relay.get_relay(app)
    await rl._on_event_message(None, 'test', {
        'key': 'testkey',
        'duration': '0s',
        'data': {'stuff': True}
    })

    await asyncio.sleep(0.01)
    q1 = asyncio.Queue()
    await rl.add_queue(q1)

    with pytest.raises(asyncio.QueueEmpty):
        q1.get_nowait()


async def test_subscribe(app, client):
    rl = relay.get_relay(app)

    await rl._on_event_message(None, 'test', {
        'key': 'testkey',
        'duration': '10s',
        'data': {'stuff': True}
    })
    strval = json.dumps({'testkey': {'stuff': True}})

    async with client.get('/sse') as resp:
        chunk = await resp.content.read(6 + len(strval))
        assert chunk.decode() == 'data: ' + strval


async def test_close(app, client):
    with pytest.raises(ClientPayloadError):
        async with client.get('/sse') as resp:
            await relay.get_relay(app).before_shutdown(app)
            await resp.content.read(5)
            await resp.content.read(5)



"""
Subscribes to eventbus messages, and forwards them to SSE clients
"""

import asyncio
import json
import weakref
from time import time
from typing import Set

from aiohttp import hdrs, web
from aiohttp_sse import sse_response
from brewblox_service import brewblox_logger, events, features, repeater, strex
from pytimeparse import parse
from schema import Optional, Schema

PUBLISH_TIMEOUT_S = 5
DEFAULT_DURATION = '60s'
CLEANUP_INTERVAL_S = 10

_message_schema = Schema({
    'key': str,
    'duration': Optional(lambda s: parse(s) is not None),
    'data': dict,
})

LOGGER = brewblox_logger(__name__)
routes = web.RouteTableDef()


def setup(app: web.Application):
    app.router.add_routes(routes)
    features.add(app, EventRelay(app))


def get_relay(app: web.Application):
    return features.get(app, EventRelay)


def _cors_headers(request):
    return {
        'Access-Control-Allow-Origin': '*',
        'Access-Control-Allow-Methods':
        request.headers.get('Access-Control-Request-Method', ','.join(hdrs.METH_ALL)),
        'Access-Control-Allow-Headers':
        request.headers.get('Access-Control-Request-Headers', '*'),
        'Access-Control-Allow-Credentials': 'true',
    }


class EventRelay(repeater.RepeaterFeature):

    def __init__(self, app: web.Application):
        super().__init__(app)
        self._queues: Set[asyncio.Queue] = weakref.WeakSet()
        self._messages: dict = {}

    def __str__(self):
        return f'<{type(self).__name__} ({len(self._queues)} listeners)>'

    async def prepare(self):
        events.get_listener(self.app).subscribe(
            exchange_name=self.app['config']['broadcast_exchange'],
            routing='#')

    async def before_shutdown(self, _):
        for queue in self._queues:
            await queue.put(asyncio.CancelledError())

    async def run(self):
        await asyncio.sleep(CLEANUP_INTERVAL_S)
        self.cleanup()

    def cleanup(self):
        now = time()
        expired = [k for k, v in self._messages.items() if v['expires'] < now]
        for key in expired:
            del self._messages[key]

    async def add_queue(self, queue: asyncio.Queue):
        self._queues.add(queue)
        try:
            self.cleanup()
            LOGGER.info(f'Added queue, setting messages: {self._messages.keys()}')
            for message in self._messages.values():
                await queue.put({message['key']: message['data']})

        except asyncio.CancelledError:  # pragma: no cover
            raise

        except Exception as ex:  # pragma: no cover
            LOGGER.info(f'Initial subscription push failed: {strex(ex)}')
            raise ex

    async def _on_event_message(self,
                                subscription: events.EventSubscription,
                                routing: str,
                                message: dict):

        _message_schema.validate(message)

        duration = parse(message.get('duration', DEFAULT_DURATION))
        message['expires'] = time() + duration

        key = message['key']
        data = message['data']
        self._messages[key] = message

        coros = [q.put({key: data}) for q in self._queues]
        await asyncio.wait_for(asyncio.gather(*coros, return_exceptions=True), PUBLISH_TIMEOUT_S)


@routes.get('/sse')
async def subscribe(request: web.Request) -> web.Response:
    """
    ---
    summary: Push all events as they are received
    tags:
    - SSE
    operationId: sse.subscribe
    produces:
    - application/json
    """
    async with sse_response(request, headers=_cors_headers(request)) as resp:
        relay: EventRelay = get_relay(request.app)
        queue = asyncio.Queue()
        await relay.add_queue(queue)

        while True:
            data = await queue.get()
            if isinstance(data, Exception):
                raise data
            await resp.send(json.dumps(data))

    # Note: we don't ever expect to return the response
    # Either the client cancels the request, or an exception is raised by publisher



"""
Tests brewblox_emitter.__main__.py
"""

from brewblox_service import events

from brewblox_emitter import __main__ as main
from brewblox_emitter import relay

TESTED = main.__name__


def test_main(mocker, app):
    mocker.patch(TESTED + '.service.run')
    mocker.patch(TESTED + '.service.create_app').return_value = app

    main.main()

    assert None not in [
        events.get_listener(app),
        relay.get_relay(app),
    ]



"""
Master file for pytest fixtures.
Any fixtures declared here are available to all test functions in this directory.
"""


import logging

import pytest
from brewblox_service import service

from brewblox_emitter.__main__ import create_parser


@pytest.fixture(scope='session', autouse=True)
def log_enabled():
    """Sets log level to DEBUG for all test functions.
    Allows all logged messages to be captured during pytest runs"""
    logging.getLogger().setLevel(logging.DEBUG)
    logging.captureWarnings(True)


@pytest.fixture
def app_config() -> dict:
    return {
        'name': 'test_app',
        'host': 'localhost',
        'port': 1234,
        'debug': False,
        'broadcast_exchange': 'brewcast_test',
    }


@pytest.fixture
def sys_args(app_config) -> list:
    return [str(v) for v in [
        'app_name',
        '--name', app_config['name'],
        '--host', app_config['host'],
        '--port', app_config['port'],
        '--broadcast-exchange', app_config['broadcast_exchange'],
    ]]


@pytest.fixture
def app(sys_args):
    parser = create_parser('default')
    app = service.create_app(parser=parser, raw_args=sys_args[1:])
    return app


@pytest.fixture
def client(app, aiohttp_client, loop):
    """Allows patching the app or aiohttp_client before yielding it.

    Any tests wishing to add custom behavior to app can override the fixture
    """
    return loop.run_until_complete(aiohttp_client(app))



"""
Example of how to import and use the brewblox service
"""

from argparse import ArgumentParser

from brewblox_service import brewblox_logger, events, scheduler, service

from brewblox_emitter import relay

LOGGER = brewblox_logger(__name__)


def create_parser(default_name='emitter') -> ArgumentParser:
    parser: ArgumentParser = service.create_parser(default_name=default_name)

    parser.add_argument('--broadcast-exchange',
                        help='Eventbus exchange to which device services broadcast their state. [%(default)s]',
                        default='brewcast_latest')

    return parser


def main():
    app = service.create_app(parser=create_parser())

    scheduler.setup(app)
    events.setup(app)
    relay.setup(app)

    service.furnish(app)
    service.run(app)


if __name__ == '__main__':
    main()
